"""Logical combinators and modifiers of schemas."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .compile import compile_ as _compile, register as _register
from .types import ValidationError

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .types import Context, Json, Path, Schema, Validator


@_register
def not_(defn: Schema, context: Context) -> Validator:
    validator = _compile(defn["not"], context)

    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        if not any(validator(x, path)):
            yield ValidationError(
                tuple(path), f"{x} should not satisfy {defn['not']}", "not"
            )

    return validate


@_register
def all_of(defn: Schema, context: Context) -> Validator:
    validators = [_compile(s, context) for s in defn["allOf"]]

    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        for v in validators:
            yield from v(x, path)

    return validate


@_register
def any_of(defn: Schema, context: Context) -> Validator:
    validators = [_compile(s, context) for s in defn["anyOf"]]

    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        messages = []
        for v in validators:
            errs = list(v(x, path))
            if not errs:
                return ()

            messages.extend(err.message for err in errs)

        failures = ", ".join(messages)
        return (
            ValidationError(
                tuple(path), f"{x} failed all conditions: {failures}", "anyOf"
            ),
        )

    return validate


@_register
def one_of(defn: Schema, context: Context) -> Validator:
    validators = [_compile(s, context) for s in defn["oneOf"]]

    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        passed = 0

        for v in validators:
            if not any(v(x, path)):
                passed += 1

        if passed != 1:
            yield ValidationError(
                tuple(path), f"{x} satisfied {passed} (!= 1) of the conditions", "oneOf"
            )

    return validate


@_register
def if_(defn: Schema, context: Context) -> Validator | None:
    if_schema = defn["if"]
    then_schema = defn.get("then", True)
    else_schema = defn.get("else", True)

    if then_schema is True and else_schema is True:
        return None

    if_validator = _compile(if_schema, context)
    then_validator = _compile(then_schema, context)
    else_validator = _compile(else_schema, context)

    def validate(x: Json, path: Path) -> Iterable[ValidationError]:
        if not any(if_validator(x, path)):  # XXX: no errors => if condition true
            yield from then_validator(x, path)
        else:
            yield from else_validator(x, path)

    return validate



"""Format validators.

Example: ...
"""

from __future__ import annotations

import ipaddress as _ipaddress
import re as _re
from datetime import date as _date, datetime as _datetime
from uuid import UUID as _UUID


def is_date_time(x: str) -> bool:
    """Date-time, see RFC 3339, section 5.6"""
    try:
        _datetime.strptime(x, "%Y-%m-%dT%H:%M:%S%z")
        return True
    except ValueError:
        return False


def is_date_time_iso(x: str) -> bool:
    """Date-time, using python's fromisoformat, rather than RFC 3339.

    Typically this is 4x faster than the spec-compliant behaviour.
    """
    try:
        _datetime.fromisoformat(x)
        return True
    except ValueError:
        return False


def is_time(x: str) -> bool:
    return is_date_time("1970-01-01T" + x)


def is_date(x: str) -> bool:
    try:
        _date.fromisoformat(x)
        return True
    except ValueError:
        return False


_EMAIL_REX = _re.compile(r"^(?!.*\.\..*@)[^@.][^@]*(?<!\.)@[^@]+\.[^@]+\Z")


def is_email(x: str) -> bool:
    """Internet email address, see RFC 5322, section 3.4.1."""
    return bool(_EMAIL_REX.fullmatch(x))


def is_ipv4(x: str) -> bool:
    """IPv4 address, see RFC 2673, section 3.2."""
    try:
        _ipaddress.IPv4Address(x)
        return True
    except ValueError:
        return False


def is_ipv6(x: str) -> bool:
    """IPv6 address, see RFC 2373, section 2.2."""
    try:
        address = _ipaddress.IPv6Address(x)
        return not getattr(address, "scope_id", "")  # ???
    except ValueError:
        return False


def is_regex(x: str) -> bool:
    try:
        _re.compile(x)
        return True
    except _re.error:
        return False


def is_uuid(x: str) -> bool:
    try:
        _UUID(x)
    except ValueError:
        return False

    # Python doesn't require seperators, RFC does.
    return all(x[position] == "-" for position in (8, 13, 18, 23))


try:
    import idna as _idna
    import jsonpointer as _jsonpointer
    import rfc3987 as _rfc3987
    import uri_template as _uri_template
    from fqdn import FQDN as _FQDN

    def is_hostname(x: str) -> bool:
        """Internet host name, see RFC 1034, section 3.1."""
        try:
            return _FQDN(x).is_valid  # pyright: ignore[reportReturnType] (this is really a bool)
        except ValueError:
            return False

    def is_idn_host_name(x: str) -> bool:
        """Internationalized domain name, see RFC 5891."""
        try:
            _idna.encode(x)
            return True
        except ValueError:
            return False

    def is_uri(x: str) -> bool:
        """A universal resource identifier (URI), see RFC3986."""
        try:
            _rfc3987.parse(x, rule="URI")
            return True
        except ValueError:
            return False

    def is_uri_reference(x: str) -> bool:
        """A URI reference, see RFC3986, section 4.1."""
        try:
            _rfc3987.parse(x, rule="URI_reference")
            return True
        except ValueError:
            return False

    def is_iri(x: str) -> bool:
        """An internationalized resource identifier (IRI), see RFC3987."""
        try:
            _rfc3987.parse(x, rule="IRI")
            return True
        except ValueError:
            return False

    def is_iri_reference(x: str) -> bool:
        """An IRI reference, see RFC3987, section 2.2."""
        try:
            _rfc3987.parse(x, rule="IRI_reference")
            return True
        except ValueError:
            return False

    def is_json_pointer(x: str) -> bool:
        """A JSON pointer, see RFC 6901."""
        try:
            return bool(_jsonpointer.JsonPointer(x))
        except _jsonpointer.JsonPointerException:
            return False

    def is_uri_template(x: str) -> bool:
        try:
            return _uri_template.validate(x)
        except ValueError:
            return False

except ImportError:
    pass


def is_relative_json_pointer(x: str) -> bool:
    """A JSON pointer relative to the current document.

    See https://tools.ietf.org/html/draft-handrews-relative-json-pointer-01#section-3
    """
    non_negative_integer, rest = [], ""
    for i, character in enumerate(x):
        if character.isdigit():
            # digits with a leading "0" are not allowed
            if i > 0 and int(x[i - 1]) == 0:
                return False

            non_negative_integer.append(character)
            continue

        if not non_negative_integer:
            return False

        rest = x[i:]
        break

    return (rest == "#") or is_json_pointer(rest)


__duration_date_order = _re.compile("^P([0-9.,]+Y)?([0-9.,]+M)?([0-9.,]+D)?$")
__duration_time_order = _re.compile("^([0-9.,]+H)?([0-9.,]+M)?([0-9.,]+S)?$")
# Only the final part of the duration can be fractional
__duration_regular = "[0-9]+[A-Z]"
__duration_final = "[0-9]+([.,][0-9]+)?[A-Z]"
__duration_fraction_order = _re.compile(
    f"^P({__duration_regular})*T?({__duration_regular})*{__duration_final}$"
)


def is_duration(x: str) -> bool:
    if len(x) < 2 or x[0] != "P":
        return False

    if not __duration_fraction_order.fullmatch(x):
        return False

    pieces = x.split("T")
    if len(pieces) == 1:
        return bool(__duration_date_order.fullmatch(x))
    elif len(pieces) == 2:
        return bool(
            pieces[1]
            and __duration_date_order.fullmatch(pieces[0])
            and __duration_time_order.fullmatch(pieces[1])
        )
    else:
        return False


FORMATS = {}
for name, obj in list(locals().items()):
    if name == "is_date_time_iso":
        continue  # not loaded by default

    # is_some_thing is the "some-thing" validator
    pieces = name.split("_")
    if pieces[0] == "is":
        key = "-".join(pieces[1:])
        FORMATS[key] = obj



from __future__ import annotations

import functools as _functools
import re as _re
from typing import TYPE_CHECKING

from .basic import _max_len_validator, _min_len_validator, _type_guard
from .compile import compile_ as _compile, register as _register
from .types import ValidationError

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import TypeVar

    from .types import Context, Json, Path, Schema, Validator

    _VT = TypeVar("_VT")

_object_guard = _functools.partial(_type_guard, schema_types=("object",), py_types=dict)


@_register
def max_properties(defn: Schema, context: Context) -> Validator:
    value: int = defn["maxProperties"]
    guard = _object_guard(defn)
    return guard(_max_len_validator(value, "maxProperties"))


@_register
def min_properties(defn: Schema, context: Context) -> Validator:
    value: int = defn["minProperties"]
    guard = _object_guard(defn)
    return guard(_min_len_validator(value, "minProperties"))


@_register
def property_names(defn: Schema, context: Context) -> Validator:
    validator = _compile(defn["propertyNames"], context)

    @_object_guard(defn)
    def validate(x: dict[str, Json], path: Path) -> ...:
        for key in x:
            yield from validator(key, path)

    return validate


@_register
def required(defn: Schema, context: Context) -> Validator | None:
    value: list[str] = defn["required"]
    if value:

        @_object_guard(defn)
        def validate(x: dict[str, Json], path: Path) -> ...:
            for v in value:
                if v not in x:
                    yield ValidationError(
                        tuple(path), f"{v} is a required property", "required"
                    )

        return validate

    return None


@_register
def dependencies(defn: Schema, context: Context) -> Validator | None:
    value: dict[str, list[str] | Schema] = defn["dependencies"]
    if not value:
        return None

    checkers = {}

    for dependent, requirement in value.items():
        if isinstance(requirement, list):
            # Has the effect of 'activating' a required directive
            fake_schema = {"required": requirement}
            if "type" in defn:
                fake_schema["type"] = defn["type"]

            checker = required(fake_schema, context)
        else:
            checker = _compile(requirement, context)

        if checker is not None:
            checkers[dependent] = checker

    @_object_guard(defn)
    def validate(x: dict[str, Json], path: Path) -> ...:
        for dependent, checker in checkers.items():
            if dependent in x:
                for err in checker(x, path):
                    yield ValidationError(
                        tuple(path),
                        f"dependency for {dependent} not satisfied: {err.message}",
                        "dependencies",
                    )

        return None

    return validate


def _path_push_iterator(
    path: list[str | int], obj: dict[str, _VT]
) -> Iterable[tuple[str, _VT]]:
    path.append("")  # front-load memory allocation
    try:
        for key, value in obj.items():
            path[-1] = key
            yield key, value
    finally:
        path.pop()


@_register
def properties(defn: Schema, context: Context) -> Validator:
    value = defn["properties"]
    validators = {k: _compile(v, context) for k, v in value.items()}

    @_object_guard(defn)
    def validate(x: dict[str, Json], path: Path) -> Iterable[ValidationError]:
        for k, v in _path_push_iterator(path, x):
            if k in validators:
                yield from validators[k](v, path)

    return validate


@_register
def pattern_properties(defn: Schema, context: Context) -> Validator:
    value = defn["patternProperties"]
    validators = [(_re.compile(k), _compile(v, context)) for k, v in value.items()]

    @_object_guard(defn)
    def validate(x: dict[str, Json], path: Path) -> Iterable[ValidationError]:
        # ugh...
        for rex, val in validators:
            for k, v in _path_push_iterator(path, x):
                if rex.search(k):
                    yield from val(v, path)

    return validate


@_register
def additional_properties(defn: Schema, context: Context) -> Validator:
    value = defn["additionalProperties"]
    simple_validator = _compile(value, context)

    excluded_names = set(defn.get("properties", ()))
    excluded_rexes = [_re.compile(k) for k in defn.get("patternProperties", ())]

    @_object_guard(defn)
    def validate(x: dict[str, Json], path: Path) -> Iterable[ValidationError]:
        for k, v in _path_push_iterator(path, x):
            if k in excluded_names:
                continue
            if any(r.match(k) for r in excluded_rexes):
                continue
            yield from simple_validator(v, path)

    return validate



"""Resolve $ref fields in schemas to actual values.

We do this when we parse the schema and cache the results.
This code is

"""

from __future__ import annotations

import contextlib as _contextlib
import json as _json
import urllib.parse as _urlparse
import urllib.request as _request
from typing import TYPE_CHECKING as _TYPE_CHECKING

from fastjsonschema.ref_resolver import (
    RefResolver as _FastRefResolver,
    get_id,
    normalize,
    resolve_path,
    resolve_remote,
)

if _TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from .types import Json, Schema, Validator


class RefTracker:
    """Tracks which identifiers have validations defined.

    This is required solely to allow existence of defered references.
    For example `{"properties": {"foo": "$ref": "#"}}` is infinitely
    recursive.
    """

    def __init__(
        self, schema: Schema | bool, handlers: dict[str, Callable[[str], Json]]
    ) -> None:
        # Trackers for various states of compilation
        self._queued: list[str] = []
        self._picked: set[str] = set()
        self.compiled: dict[str, Validator] = {}

        self._resolver = RefResolver.from_schema(schema, store={}, handlers=handlers)

        # Kick off the compilation with top-level function
        self._queued.append(self._resolver.get_uri())
        self._entrypoint_uri = self._queued[0]

    def __bool__(self) -> bool:
        return bool(self._queued)

    def queue(self, uri: str) -> None:
        self._queued.append(uri)

    def pop(self) -> str:
        uri = self._queued.pop()
        self._picked.add(uri)
        return uri

    def seen(self, uri: str) -> bool:
        return uri in self._picked

    @property
    def entrypoint(self) -> Validator:
        return self.compiled[self._entrypoint_uri]


def request_handler(uri: str) -> Json:
    request = _request.Request(uri, headers={"User-Agent": "jsonscreamer"})  # noqa: S310
    with _request.urlopen(request) as resp:  # noqa: S310
        return _json.load(resp)


HANDLERS = {"http": request_handler, "https": request_handler}


class RefResolver(_FastRefResolver):
    """
    Resolve JSON References.

    This is a partial rewrite of fastjsonschema's implementation,
    with some compatibility hacks. See:
        https://github.com/horejsek/python-fastjsonschema
    """

    _TRAVERSE_ARBITRARY_KEYS = frozenset(("definitions", "properties"))

    def __init__(
        self,
        base_uri: str,
        schema: Schema | bool,
        store: object = ...,
        cache: bool = True,
        handlers: object = ...,
    ) -> None:
        # XXX: import here must be deferred to prevent cyclic imports
        from .compile import _COMPILATION_FUNCTIONS

        self._TRAVERSABLE_KEYS = (
            frozenset(_COMPILATION_FUNCTIONS)
            .union(("definitions",))
            .difference(("const", "enum"))
        )

        super().__init__(base_uri, schema, store, cache, handlers)

    @_contextlib.contextmanager
    def resolving(self, ref: str) -> Iterator[Schema | bool]:
        """
        Context manager which resolves a JSON ``ref`` and enters the
        resolution scope of this ref.
        """
        absolute = bool(_urlparse.urlsplit(ref).netloc)

        new_uri = ref if absolute else _urlparse.urljoin(self.resolution_scope, ref)
        uri, fragment = _urlparse.urldefrag(new_uri)

        # TODO: edge case - fragments in ids - remove for later schemas
        if new_uri and new_uri in self.store:
            schema: Schema = self.store[new_uri]
            fragment = ""
        elif uri and normalize(uri) in self.store:
            schema = self.store[normalize(uri)]
        elif not uri or uri == self.base_uri:
            schema = self.schema
        else:
            schema = resolve_remote(uri, self.handlers)
            if self.cache:
                self.store[normalize(uri)] = schema

        old_base_uri, old_schema = self.base_uri, self.schema
        self.base_uri, self.schema = uri, schema
        try:
            with self.in_scope(uri):
                yield resolve_path(schema, fragment)
        finally:
            self.base_uri, self.schema = old_base_uri, old_schema

    def walk(self, node: dict, arbitrary_keys: bool = False) -> None:
        """
        Walk thru schema and dereferencing ``id`` and ``$ref`` instances
        """
        if isinstance(node, bool):
            pass
        elif "$ref" in node and isinstance(node["$ref"], str):
            ref = node["$ref"]
            node["$ref"] = _urlparse.urljoin(self.resolution_scope, ref)
        elif ("$id" in node or "id" in node) and isinstance(get_id(node), str):
            with self.in_scope(get_id(node)):
                self.store[normalize(self.resolution_scope)] = node
                # TODO: edge case - fragments in ids - remove for later schemas
                self.store[self.resolution_scope] = node
                for key, item in node.items():
                    if isinstance(item, dict) and (
                        arbitrary_keys or key in self._TRAVERSABLE_KEYS
                    ):
                        self.walk(
                            item, arbitrary_keys=key in self._TRAVERSE_ARBITRARY_KEYS
                        )
        else:
            for key, item in node.items():
                if isinstance(item, dict) and (
                    arbitrary_keys or key in self._TRAVERSABLE_KEYS
                ):
                    self.walk(item, arbitrary_keys=key in self._TRAVERSE_ARBITRARY_KEYS)



from __future__ import annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING

from .types import ValidationError

if _TYPE_CHECKING:
    from collections.abc import Callable, Iterable
    from typing import TypeVar

    from .types import Compiler, Context, Json, Path, Schema, Validator

    _CT = TypeVar("_CT", bound=Compiler)


_COMPILATION_FUNCTIONS: dict[str, Compiler] = {}
_TYPE_CHECKERS = {
    "object": lambda x: isinstance(x, dict),
    "array": lambda x: isinstance(x, list),
    "string": lambda x: isinstance(x, str),
    "number": lambda x: (isinstance(x, (float, int)) and not isinstance(x, bool)),
    "integer": lambda x: (
        (isinstance(x, int) and not isinstance(x, bool))
        or (isinstance(x, float) and x == int(x))
    ),
    "boolean": lambda x: isinstance(x, bool),
    "null": lambda x: x is None,
}


def type_(
    defn: Schema, context: Context
) -> Callable[[Json, Path], ValidationError | None]:
    """Create a validator to check the type of an item.

    Unlike other validators we do not yield a value, but return it, since we
    want to break early and not run the other validations if they type check fails.
    """
    required_type: str | list[str] = defn["type"]

    if isinstance(required_type, str):
        type_checker = _TYPE_CHECKERS[required_type]

        def validate(x: Json, path: Path) -> ValidationError | None:
            if not type_checker(x):
                return ValidationError(
                    tuple(path), f"{x} is not of type '{required_type}'", "type"
                )

    elif isinstance(required_type, list):
        type_checkers = [_TYPE_CHECKERS[v] for v in required_type]

        def validate(x: Json, path: Path) -> ValidationError | None:
            if not any(t(x) for t in type_checkers):
                return ValidationError(
                    tuple(path),
                    f"{x} is not any of the types '{required_type}'",
                    "type",
                )

    return validate


def register(validator: _CT) -> _CT:
    """Register a validator for compiling a given type."""
    _COMPILATION_FUNCTIONS[_name_from_validator(validator)] = validator
    return validator


def compile_(defn: Schema | bool, context: Context) -> Validator:
    if defn is True or defn == {}:
        return _true
    elif defn is False:
        return _false
    elif not isinstance(defn, dict):
        raise ValueError("definition must be a boolean or object")

    if "$ref" in defn:
        validate = compile_ref(defn, context)

    else:
        type_validator = None
        validators = []
        for key in defn:
            if key == "type":
                type_validator = type_(defn, context)
            elif key in _COMPILATION_FUNCTIONS:
                validator = _COMPILATION_FUNCTIONS[key](defn, context)
                if validator is not None:
                    validators.append(validator)

        if type_validator:

            def validate(x: Json, path: Path) -> Iterable[ValidationError]:
                if err := type_validator(x, path):
                    yield err
                else:
                    for v in validators:
                        yield from v(x, path)

        else:

            def validate(x: Json, path: Path) -> Iterable[ValidationError]:
                for v in validators:
                    yield from v(x, path)

    return validate


def compile_ref(defn: Schema, context: Context) -> Validator:
    tracker = context.tracker

    with tracker._resolver.in_scope(defn["$ref"]):
        uri = tracker._resolver.get_uri()
        if uri not in tracker._picked:
            tracker.queue(uri)

        def validate(x: Json, path: Path) -> Iterable[ValidationError]:
            yield from tracker.compiled[uri](x, path)

        return validate


def _name_from_validator(validator: Callable) -> str:
    pieces = validator.__name__.strip("_").split("_")
    # JSON Scheam uses camelCase
    for ix, piece in enumerate(pieces[1:]):
        pieces[ix + 1] = piece.capitalize()

    return "".join(pieces)


def _true(x: Json, path: Path) -> tuple[ValidationError, ...]:
    return ()


def _false(x: Json, path: Path) -> tuple[ValidationError]:
    return (ValidationError(tuple(path), f"{x} cannot satisfy false", "false"),)



from __future__ import annotations

import functools
from typing import TYPE_CHECKING as _TYPE_CHECKING

from .basic import (
    _max_len_validator,
    _min_len_validator,
    _strict_bool_nested,
    _type_guard,
)
from .compile import compile_ as _compile, register as _register
from .types import ValidationError

if _TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import TypeVar

    from .types import Context, Json, Path, Schema, Validator

    _T = TypeVar("_T")


_array_guard = functools.partial(_type_guard, schema_types=("array",), py_types=list)


@_register
def min_items(defn: Schema, context: Context) -> Validator | None:
    value: int = defn["minItems"]
    guard = _array_guard(defn)
    return guard(_min_len_validator(value, "minItems"))


@_register
def max_items(defn: Schema, context: Context) -> Validator | None:
    value: int = defn["maxItems"]
    guard = _array_guard(defn)
    return guard(_max_len_validator(value, "maxItems"))


def _unique_checker(
    x: Json, path: list[str | int], _second_run: bool = False
) -> Iterable[ValidationError]:
    # Happy path first, then fall back to more expensive expressions
    if _second_run:
        x = _strict_bool_nested(x)  # type: ignore

    ok = True
    try:
        ok = len(x) == len(set(x))  # type: ignore (guarded)
    except TypeError:
        # No choice but to fall back to O(n^2) algorithm
        seen = []
        for item in x:  # type: ignore (guarded)
            if item in seen:
                ok = False
                break
            seen.append(item)

    if not ok and not _second_run:
        # Edge case: both booleans and integers in the array
        # checking this false positive will be slower..
        ok = not any(_unique_checker(x, path, _second_run=True))

    if not ok:
        yield ValidationError(tuple(path), f"{x} has repeated items", "uniqueItems")


@_register
def unique_items(defn: Schema, context: Context) -> Validator | None:
    if defn["uniqueItems"]:
        guard = _array_guard(defn)
        return guard(_unique_checker)


def _path_push_iterator(
    path: Path, iterable: Iterable[_T], offset: int = 0
) -> Iterable[_T]:
    path.append(0)  # front-load memory allocation
    try:
        for ix, item in enumerate(iterable):
            path[-1] = ix + offset
            yield item
    finally:
        path.pop()


@_register
def items(defn: Schema, context: Context) -> Validator | None:
    value: list[Schema] | Schema = defn["items"]
    if isinstance(value, list):
        validators = [_compile(d, context) for d in value]

        @_array_guard(defn)
        def validate(x: list[Json], path: Path) -> Iterable[ValidationError]:
            for v, i in _path_push_iterator(path, zip(validators, x)):
                yield from v(i, path)

    else:
        validator = _compile(value, context)

        @_array_guard(defn)
        def validate(x: list[Json], path: Path) -> Iterable[ValidationError]:
            for i in _path_push_iterator(path, x):
                yield from validator(i, path)

    return validate


@_register
def additional_items(defn: Schema, context: Context) -> Validator | None:
    item_spec: dict | list = defn.get("items", {})
    if isinstance(item_spec, dict):
        # this is a no-op
        return None

    offset = len(item_spec)
    validator = _compile(defn["additionalItems"], context)

    @_array_guard(defn)
    def validate(x: list[Json], path: Path) -> Iterable[ValidationError]:
        for i in _path_push_iterator(path, x[offset:], offset):
            yield from validator(i, path)

    return validate


@_register
def contains(defn: Schema, context: Context) -> Validator | None:
    validator = _compile(defn["contains"], context)
    if validator is None:
        return None

    @_array_guard(defn)
    def validate(x: list[Json], path: Path) -> Iterable[ValidationError]:
        for i in x:
            if not any(validator(i, path)):  # XXX: no error => item is in list
                return ()

        return (
            ValidationError(
                tuple(path),
                f"{x} did not contain any items satisfying {defn['contains']}",
                "contains",
            ),
        )

    return validate


@_register
def max_contains(defn: Schema, context: Context) -> Validator | None:
    if "contains" not in defn:
        return None

    validator = _compile(defn["contains"], context)
    value: int = defn["maxContains"]

    @_array_guard(defn)
    def validate(x: list[Json], path: Path) -> Iterable[ValidationError]:
        total = sum(not any(validator(i, path)) for i in x)
        if total <= value:
            return ()

        return (
            ValidationError(
                tuple(path),
                f"{x} contains more than {value} items satisfying {defn['contains']}",
                "maxContains",
            ),
        )

    return validate


@_register
def min_contains(defn: Schema, context: Context) -> Validator | None:
    if "contains" not in defn:
        return None

    validator = _compile(defn["contains"], context)
    value: int = defn["minContains"]

    @_array_guard(defn)
    def validate(x: list[Json], path: Path) -> Iterable[ValidationError]:
        total = sum(not any(validator(i, path)) for i in x)
        if total >= value:
            return ()

        return (
            ValidationError(
                tuple(path),
                f"{x} contains less than {value} items satisfying {defn['contains']}",
                "minContains",
            ),
        )

    return validate



from __future__ import annotations

from collections.abc import (
    Callable as _Callable,
    Mapping as _Mapping,
    Sequence as _Sequence,
)
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any as _Any, Protocol as _Protocol

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .resolve import RefTracker

Json = None | bool | int | float | str | _Sequence["Json"] | _Mapping[str, "Json"]
Path = list[str | int]
ImmutablePath = tuple[str | int, ...]
Format = _Callable[[str], bool]


class ValidationError(ValueError):
    """Raised when an instance does not conform to the provided schema."""

    def __init__(self, absolute_path: ImmutablePath, message: str, type: str) -> None:
        self.absolute_path = absolute_path
        self.message = message
        self.type = type


@dataclass
class Context:
    formats: dict[str, Format]
    tracker: RefTracker


Schema = dict[str, _Any]
Result = ValidationError | None


class Validator(_Protocol):
    def __call__(self, x: Json, path: Path) -> Iterable[ValidationError]: ...


Compiler = _Callable[[Schema, Context], Validator | None]
